#include <stdatomic.h>
#include <unrelacy.h>
#include <common.h>

#include <spec_lib.h>
#include <stdlib.h>
#include <cdsannotate.h>
#include <specannotation.h>
#include <model_memory.h>
#include "common.h" 

/**
	@Begin
	@Class_begin
	@End
*/
template <typename t_element, size_t t_size>
struct mpmc_boundq_1_alt
{
private:

	// elements should generally be cache-line-size padded :
	t_element               m_array[t_size];

	// rdwr counts the reads & writes that have started
	atomic<unsigned int>    m_rdwr;
	// "read" and "written" count the number completed
	atomic<unsigned int>    m_read;
	atomic<unsigned int>    m_written;

public:

	mpmc_boundq_1_alt()
	{
    	/**
	    	@Begin
			@Entry_point
			@End
		*/
		m_rdwr = 0;
		m_read = 0;
		m_written = 0;
	}
	

	/**
		@Begin
		@Options:
			LANG = CPP;
			CLASS = mpmc_boundq_1_alt;
		@Global_define:
		@DeclareVar:
			id_tag_t *tag;
		@InitVar:
			tag = NULL;
	@Happens_before:
		Publish -> Fetch
	@End
	*/

	//-----------------------------------------------------

	/**
		@Begin
		@Interface: Fetch
		@Commit_point_set: Fetch_Succ_Point | Fetch_Fail_Point
		@ID: (call_id_t) __RET__
		@End
	*/
	t_element * read_fetch() {
		unsigned int rdwr = m_rdwr.load(mo_acquire);
		/**
			@Begin
			@Commit_point_define_check: (unsigned int) ((rdwr>>16) & 0xFFFF) == (unsigned int) (rdwr & 0xFFFF)
			@Label: Fetch_Fail_Point
			@End
		*/
		unsigned int rd,wr;
		for(;;) {
			rd = (rdwr>>16) & 0xFFFF;
			wr = rdwr & 0xFFFF;

			//model_print("cond: %d\n", (unsigned int) ((rdwr>>16) & 0xFFFF) ==
			//(unsigned int) (rdwr & 0xFFFF));
			//model_print("cond: %d\n", wr == rd);
			if ( wr == rd ) { // empty
				/**
					//@Begin
					@Commit_point_define: true
					@Potential_commit_point_label: Fetch_Potential_Point 
					@Label: Fetch_Fail_Point
					@End
				*/
				model_print("in cond\n");
				return false;
			}
			
			bool succ = m_rdwr.compare_exchange_weak(rdwr,rdwr+(1<<16),mo_acq_rel);
			
			if (succ)
				break;
			else
				thrd_yield();
		}

		// (*1)
		rl::backoff bo;
		
		while ( true ) {
			if ((m_written.load(mo_acquire) & 0xFFFF) == wr) {
				/**
					@Begin
					@Commit_point_define_check: true 
					@Label: Fetch_Succ_Point
					@End
				*/
				break;
			}
			thrd_yield();
		}

		t_element * p = & ( m_array[ rd % t_size ] );
		
		return p;
	}

	void read_consume() {
		m_read.fetch_add(1,mo_release);
	}

	//-----------------------------------------------------

	t_element * write_prepare() {
		unsigned int rdwr = m_rdwr.load(mo_acquire);
		unsigned int rd,wr;
		for(;;) {
			rd = (rdwr>>16) & 0xFFFF;
			wr = rdwr & 0xFFFF;

			if ( wr == ((rd + t_size)&0xFFFF) ) { // full
				return NULL;
			}
			
			bool succ = m_rdwr.compare_exchange_weak(rdwr,(rd<<16) |
				((wr+1)&0xFFFF),mo_acq_rel);
			if (succ)
				break;
			else
				thrd_yield();
		}

		// (*1)
		rl::backoff bo;
		while ( (m_read.load(mo_acquire) & 0xFFFF) != rd ) {
			thrd_yield();
		}

		t_element * p = & ( m_array[ wr % t_size ] );

		return p;
	}

	/**
		@Begin
		@Interface: Publish 
		@Commit_point_set: Publish_Point
		@ID: (uint64_t) elem
		@End
	*/
	void write_publish(t_element *elem)
	{
		m_written.fetch_add(1,mo_release);
		/**
			@Begin
			@Commit_point_define_check: true
			@Label: Publish_Point
			@End
		*/
	}

	//-----------------------------------------------------


};
/**
	@Begin
	@Class_end
	@End
*/
