#include <atomic>
#include <threads.h>
#include <stdatomic.h>
#include <stdlib.h>
#include <stdio.h>

#include <spec_lib.h>
#include <stdlib.h>
#include <cdsannotate.h>
#include <specannotation.h>
#include <model_memory.h>
#include "common.h" 

#include "librace.h"

/**
	This is an example about how to specify the correctness of the
	read-copy-update synchronization mechanism.
*/

typedef struct Data {
	int data1;
	int data2;
	int data3;
} Data;

atomic<Data*> data;
	
/**
	@Begin
	@Global_define:
		@DeclareVar:
			Data *_cur_data;
		@InitVar:
			_cur_data = NULL;
		
		@DefineFunc:
		bool equals(Data *ptr1, Data *ptr2) {
			//if (ptr1->data1 == ptr2->data1
			//	&& ptr1->data2 == ptr2->data2
			//	&& ptr1->data3 == ptr2->data3) {
			//	return true;
			//} else {
			//	return false;
			//}
			return ptr1 == ptr2;
		}

		bool isOriginalRead(Data *ptr) {
			return ptr->data1 == 0 &&
				ptr->data2 == 0 && ptr->data3 == 0;
		}

	@Happens_before:
		Write -> Read
		Write -> Write
	@End
*/

/**
	@Begin
	@Interface: Read
	@Commit_point_set: Read_Success_Point
	//@ID: isOriginalRead(__RET__) ? 1 : DEFAULT_CALL_ID
	@Action:
		Data *_Old_Data = _cur_data;
	@Post_check:
		_Old_Data = __RET__
	@End
*/
Data* read() {
	Data *res = data.load(memory_order_acquire);
	/**
		@Begin
		@Commit_point_define_check: true
		@Label: Read_Success_Point
		@End
	*/
	//model_print("Read: %d, %d, %d\n", res->data1, res->data2, res->data3);
	return res;
}

/**
	@Begin
	@Interface: Write
	@Commit_point_set: Write_Success_Point
	@Action:
		Data *_Old_data = _cur_data;
		_cur_data = __RET__;
	@Post_check:
		_Old_data == NULL ?
			(__RET__->data1 == d1	&&
			__RET__->data2 == d2	&&
			__RET__->data3 == d3)
			:
			(__RET__->data1 == _Old_data->data1 + d1 &&
			__RET__->data2 == _Old_data->data2 + d2	&&
			__RET__->data3 == _Old_data->data3 + d3)
	@End
*/
Data* write(int d1, int d2, int d3) {
	Data *prev = data.load(memory_order_acquire);
	bool succ = false;
	Data *tmp = (Data*) malloc(sizeof(Data));
	do {
        tmp->data1 = prev->data1 + d1;
	    tmp->data2 = prev->data2 + d2;
	    tmp->data3 = prev->data3 + d3;
        succ = data.compare_exchange_strong(prev, tmp,
            memory_order_acq_rel, memory_order_acquire);
		/**
			@Begin
			@Commit_point_define_check: succ
			@Label: Write_Success_Point
			@End
		*/
	} while (!succ);
	//model_print("Write: %d, %d, %d\n", tmp->data1, tmp->data2, tmp->data3);
	return tmp;
}


void threadA(void *arg) {
	write(1, 0, 0);
}

void threadB(void *arg) {
	write(0, 1, 0);
}

void threadC(void *arg) {
	write(0, 0, 1);
}

void threadD(void *arg) {
	Data *dataC = read();
}

int user_main(int argc, char **argv) {
	/**
		@Begin
		@Entry_point
		@End
	*/
	
	thrd_t t1, t2, t3, t4;
	Data *dataInit = (Data*) malloc(sizeof(Data));
	dataInit->data1 = 0;
	dataInit->data2 = 0;
	dataInit->data3 = 0;
	atomic_init(&data, dataInit);
	write(0, 0, 0);

	thrd_create(&t1, threadA, NULL);
	thrd_create(&t2, threadB, NULL);
	//thrd_create(&t3, threadC, NULL);
	thrd_create(&t4, threadD, NULL);

	thrd_join(t1);
	thrd_join(t2);
	//thrd_join(t3);
	thrd_join(t4);

	return 0;
}
